/* ===================================
Bodenfeuchte230121.ino
Programm für ESP8266 mit kapazitivem Bodenfeuchtesensor
Autoren: Janek Westphal, Claus Brell
Stand: 19.02.2023

Zum Einsatz im Rahmen der Schulkooperationen
 des FB Wirtschaftswissenschaften
Hochschule Niederhein

Quellen:
https auf ESP8266: https://maakbaas.com/esp8266-iot-framework/logs/https-requests/
(Das funktioniert nur halb)
https://stackoverflow.com/questions/64845625/how-to-send-https-get-request-using-nodemcu
(Das funktioniert nur halb)
https://buger.dread.cz/simple-esp8266-https-client-without-verification-of-certificate-fingerprint.html
(Das funktioniert)

Ablauf:
Aufbau der WLAN-Verbindung, messen und Übertragung an den Webserver 
in loop(), damit ein Ausfall des WLANs nicht den ESP stoppt.

ESP8266 Node-MCU
Pinbelegung
              Antenne
FeuD---ADC   A0 | D0  GPIO16---LED
            RSV | D1  GPIO5
            RSV | D1  GPIO4
    GPIO10  SD3 | D1  GPIO0
    GPIO9   SD2 | D4  GPIO2 
            SD1 | 3V3 
            CMD | GND
            SD0 | D5  GPIO14
            CLK | D6  GPIO12
FeuG--------GND | D7  GPIO13
Feu+--------3V3 | D8  GPIO15
             EN | RX  GPIO3
            RST | TX  GPIO1
            GND | GND
            Vin | 3V3
      (RST)   UBatt   (FLASH)


ESP8266 D1 Mini
Pinbelegung
                  Antenne
                RST | TX  
Feu(analog)     A0  | RX  
        GPIO16  D0  | D1  GPIO5  LED+
        GPIO14  D5  | D2  GPIO4
        GPIO12  D6  | D3  GPIO0 
        GPIO13  D7  | D4  GPIO2 
        GPIO15  D8  | GND.       LED-, Feu(-)
Feu(+)          3V3 | 5V
          (RESET) UBatt   
========================================*/

// Bibliotheken für den Plattform-Zugriff
// WLAN
// Für ESP32 (auskommentiert)
//    #include <WiFi.h>
// Für Node MCU
#include <ESP8266WiFi.h>
// einfacher Zugriff auf den Wewbserver
//#include <WiFiClientSecure.h>
#include <ESP8266HTTPClient.h>
// ESP32 Wroom-Modul: Anschluss G35 = GPIO5 ist der verfügbare ADC1
// #define SENSOR_PIN 35
// ESP8266 Node MCU A0
#define SENSOR_PIN A0
// Nutzerbeglückung
// ESP32 LED an 32
// ESP8266 Node MCU SD03 oder D0=GPIO16
// ESP2866 D1Mini (AZ Delivery) D1=GPIO5 
// #define LED 16 // ESP32
#define LED 5 // D1 Mini
// Netzwerk Parameter
//const char* ssid     = "iot";
//const char* password = ":)";
//const char* ssid     = "Gloryhammer";
//const char* password = ":)";
const char* ssid     = "raspilab-1";
const char* password = "raspilab1234";
// Webserver Parameter Biene40 Plattform
const char* host = "210301.de"; // Serveradresse, ohne "https://...."
String urlpath = "/b40/WS/msg220304.php?x="; // Webservice des Biene40-Projektes
const int httpsPort=443; //default für https, 80 für http
// Parameter für die Biene40 Plattform
// für jeden Sensor unterschiedlich, hier ein Dummy Sensor
// Die richtige Kennung wird im Workshop mitgeteilt
// String SENSOR_NR   = "abc123XY";
// String UPLOAD_CODE = "abc123XY";
String SENSOR_NR   = "10000001";
String UPLOAD_CODE = "10000001";


void setup()
{
    // Set pin mode
    pinMode(LED,OUTPUT);
    // lang (1s) leuchten zum Neustart
    blinky(1,1000);

    Serial.begin(9600);
    delay(100);
    Serial.print("Start");
    Serial.println(" Ver. 230219-2050");
}


void loop()
{
    /*  -------------------------------------------
        Schritt 1: WLAN Verbindung herstellen
        -------------------------------------------*/
    WiFi.mode(WIFI_OFF);
    WiFi.mode(WIFI_AP_STA);
    if (WiFi.status() == WL_CONNECTED) Serial.println("WLAN war verbunden");
    WiFi.begin(ssid, password);
    int i=0; // Es wird nur 20 mal versucht, eine WLAN Verbindung aufzubauen
    while (!(WiFi.status() == WL_CONNECTED) && i<20) {
        blinky(1,32);
        Serial.print(i);
        i++;
    }
    Serial.println();
    if (WiFi.status() != WL_CONNECTED) Serial.println("Keine WLAN-Verbindung möglich.");
    /*  -------------------------------------------
        Nun sollte WLAN verbunden sein
        -------------------------------------------*/
    Serial.println();
    Serial.print("meine IP: ");
    Serial.println(WiFi.localIP());

    /*  -------------------------------------------
      Schritt 2: Messwert holen
      -------------------------------------------*/
    // Testsystem Bodenfeuchte Brell
    int measure = analogRead(SENSOR_PIN);
    // Skalierung auf maximal 0 bis 100
    int percent = convertMeasureToPercent(measure);

    /*  -------------------------------------------
      Schritt 3: Aufruf Webserver
      -------------------------------------------*/
    // Parameter für http-Aufruf zusammenmstellen
    // API für die Biene40-Plattform
    // Das ist der zentrale Teil
    String url = urlpath+SENSOR_NR+";"+UPLOAD_CODE+";1673815507;-10;"+percent;
    Serial.print("Aufruf: ");
    Serial.println(url);
    // Aufruf:
    callhttps(url);

    /*  -------------------------------------------
        Schrittt 4: WLAN Verbindung kappen
        und Ressourcen freigeben
        -------------------------------------------*/
    if (WiFi.status() == WL_CONNECTED) Serial.println("WLAN ist verbunden");
    WiFi.mode(WIFI_OFF);
    if (WiFi.status() != WL_CONNECTED) Serial.println("WLAN ist getrennt");
    
    /*  -------------------------------------------
        Schritt 5: Pause zwischen den Messungen
        12*15=180 mal blitzt die LED alle 5 Sekunden
        sind 180*5=900 Sekunden (15 Minuten) plus ein Bisschen
        Für den Test / Entwicklung:
        3*5=15 Sekunden
        -------------------------------------------*/
    Serial.println("===================\r\nPause ");
    // for (int i=0;i<180;i++){ // Produktion
    for (int i=0;i<3;i++){ // Test&Entwicklung
      blinky(1,22);
      Serial.print('.'); 
      delay(5000); // Pause 5000 ms = 5 s   
    } // Ende for
}

/* ------------------------------
   eigene Funktionen
   ------------------------------*/

/* -------------------------------
   Funktion:
   convertMeasureToPercent(Messwert)
   -------------------------------*/
// Messwert in den Bereich 0-100 Transformieren
int convertMeasureToPercent(int measure){
  /* Es wird von einem 10 Bit Wert ausgegangen (ESP8266), 12 Bit ESP32
     0 entstricht 100%
     1023 entspricht 0% ESP8266 (10 Bit)
     4095 entspricht 0% ESP32 (12 Bit)
     Dazwischen wird linear umgerechnet.
     Das ist nicht präzise, genügt aber für den
     Anwendungsfall Bodenfeuchtemessung
     Bislang keinen Wert über 80 und unter 20 beobachtet, daher Spreizung der Skala
     20 und kleiner -> 0%
     80 und größer -> 100%
  */
  // float quotient=1-(float)measure/4095.0; // ESP32
  float quotient=1-(float)measure/1023.0; // ESP8266
  Serial.print("Rohwert: ");
  Serial.println(measure);  
  // WARTUNG
  float skaliert=(quotient*100.0-50.0)*(100.0/(77.0-50.0));
  // Serial.print("transformierter Messwert: ");
  Serial.print("transformierter Wert: ");
  Serial.println(skaliert);  
  int percent=(int)(skaliert+0.5);
  return percent;   
} // Ende Funktion


/* -------------------------------
   Funktion:
   blinky(Anzahl, Dauer LED an)
   Rückmeldung für den Nutzer
   -------------------------------*/
void blinky(int anzahl, int dauer){
  /* Die interne LED wird zur Signalisierung genutzt
     1 Sekunde - alles ok
     64 ms Blitz - ich arbeite
     5 * halbe Sekunde - etwas Schlimmes ist passiert
  */
  for (int i=0;i<anzahl;i++){
    delay(500);
    digitalWrite(LED,HIGH);
    delay(dauer);
    digitalWrite(LED,LOW);
  } // Ende for   
} // Ende Funktion blinky ----------------------

/* -------------------------------
   Funktion:
   callhttps()
   Übergabe der gemessenen Daten an den Server
   Quelle: https://www.aeq-web.com/nodemcu-esp8266-ssl-secure-https-post-and-get-request/
   -------------------------------*/
void callhttps(String url){
  int httpsCode=666;
  BearSSL::WiFiClientSecure client;
  client.setInsecure(); //the magic line, use with caution
  HTTPClient https;
  Serial.println("Versuch Verbindungsaufbau Webservice");
  if (https.begin(client, host, httpsPort, url)) {
        httpsCode = https.GET();
        if (httpsCode > 0) {
            Serial.print("Code: ");
            Serial.println(httpsCode);
            if (httpsCode == HTTP_CODE_OK) {
                Serial.println(" :) Verbunden");
                blinky(2,500);
                Serial.println(https.getString());
            }
        } else {
            Serial.println(" :( keine Datenübertragung");
        }
    } else {
        Serial.println(" :( keine Verbindung zum Server ");
        Serial.println(httpsCode);
    }

} // Ende Funktion ------------------------------
